@@ -16,6 +16,7 @@ use std::{
 	borrow::Cow,
 	cell::RefCell,
 	collections::{hash_map::Entry, HashMap},
+	path::PathBuf,
 	sync::{
 		atomic::{AtomicBool, Ordering},
 		Arc,
@@ -44,7 +45,7 @@ use x11rb::{
 
 #[cfg(feature = "image-data")]
 use super::encode_as_png;
-use super::{into_unknown, LinuxClipboardKind, WaitConfig};
+use super::{into_unknown, paths_from_uri_list, LinuxClipboardKind, WaitConfig};
 #[cfg(feature = "image-data")]
 use crate::ImageData;
 use crate::{common::ScopeGuard, Error};
@@ -77,6 +78,7 @@ x11rb::atom_manager! {
 		TEXT_MIME_UNKNOWN: b"text/plain",
 
 		HTML: b"text/html",
+		URI_LIST: b"text/uri-list",
 
 		PNG_MIME: b"image/png",
 
@@ -134,16 +136,10 @@ impl XContext {
 	fn new() -> Result<Self> {
 		// create a new connection to an X11 server
 		let (conn, screen_num): (RustConnection, _) =
-			RustConnection::connect(None).map_err(|_| Error::Unknown {
-				description: String::from(
-					"X11 server connection timed out because it was unreachable",
-				),
+			RustConnection::connect(None).map_err(|_| {
+				Error::unknown("X11 server connection timed out because it was unreachable")
 			})?;
-		let screen = conn
-			.setup()
-			.roots
-			.get(screen_num)
-			.ok_or(Error::Unknown { description: String::from("no screen found") })?;
+		let screen = conn.setup().roots.get(screen_num).ok_or(Error::unknown("no screen found"))?;
 		let win_id = conn.generate_id().map_err(into_unknown)?;
 
 		let event_mask =
@@ -225,9 +221,7 @@ impl Inner {
 		wait: WaitConfig,
 	) -> Result<()> {
 		if self.serve_stopped.load(Ordering::Relaxed) {
-			return Err(Error::Unknown {
-                description: "The clipboard handler thread seems to have stopped. Logging messages may reveal the cause. (See the `log` crate.)".into()
-            });
+			return Err(Error::unknown("The clipboard handler thread seems to have stopped. Logging messages may reveal the cause. (See the `log` crate.)"));
 		}
 
 		let server_win = self.server.win_id;
@@ -527,9 +521,7 @@ impl Inner {
 			Ok(ReadSelNotifyResult::IncrStarted)
 		} else {
 			// this should never happen, we have sent a request only for supported types
-			Err(Error::Unknown {
-				description: String::from("incorrect type received from clipboard"),
-			})
+			Err(Error::unknown("incorrect type received from clipboard"))
 		}
 	}
 
@@ -714,9 +706,7 @@ impl Inner {
 			return Ok(());
 		}
 
-		Err(Error::Unknown {
-			description: "The handover was not finished and the condvar didn't time out, yet the condvar wait ended. This should be unreachable.".into()
-		})
+		Err(Error::unknown("The handover was not finished and the condvar didn't time out, yet the condvar wait ended. This should be unreachable."))
 	}
 }
 
@@ -885,6 +875,12 @@ impl Clipboard {
 		self.inner.write(data, selection, wait)
 	}
 
+	pub(crate) fn get_html(&self, selection: LinuxClipboardKind) -> Result<String> {
+		let formats = [self.inner.atoms.HTML];
+		let result = self.inner.read(&formats, selection)?;
+		String::from_utf8(result.bytes).map_err(|_| Error::ConversionFailure)
+	}
+
 	pub(crate) fn set_html(
 		&self,
 		html: Cow<'_, str>,
@@ -935,6 +931,14 @@ impl Clipboard {
 		let data = vec![ClipboardData { bytes: encoded, format: self.inner.atoms.PNG_MIME }];
 		self.inner.write(data, selection, wait)
 	}
+
+	pub(crate) fn get_file_list(&self, selection: LinuxClipboardKind) -> Result<Vec<PathBuf>> {
+		let result = self.inner.read(&[self.inner.atoms.URI_LIST], selection)?;
+
+		String::from_utf8(result.bytes)
+			.map_err(|_| Error::ConversionFailure)
+			.map(paths_from_uri_list)
+	}
 }
 
 impl Drop for Clipboard {
